import typing as t
from pydantic import BaseModel, HttpUrl, field_validator, IPvAnyAddress, Field
from datetime import datetime


class Event(BaseModel):
    name: str = Field(max_length=32)
    siteId: str = Field(max_length=16)
    sourceIp: IPvAnyAddress
    time: str

    @field_validator('time')
    @classmethod
    def validate_time(cls, time: str):
        datetime.strptime(time, "%Y-%m-%dT%H:%M:%SZ")
        return time

    @property
    def py_time(self):
        return datetime.strptime(self.time, "%Y-%m-%dT%H:%M:%SZ")


class PageViewEvent(Event):
    name: t.Literal['PageView']
    pageUrl: HttpUrl
    referrer: str = ""
    prevUrl: str = ""



import gunicorn.app.base


class GunicornApplication(gunicorn.app.base.BaseApplication):

    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application



import duckdb as db
import datapad as dp
from tabulate import tabulate
from datetime import datetime, timedelta, timezone
from argparse import ArgumentParser

if __name__ == "__main__":
    ap = ArgumentParser()
    ap.add_argument("-l", "--log", default="EventLog.jsonl.gz",
                    help="path to where we should save event log file")

    args = ap.parse_args()
    event_log_file = args.log

    stmt = \
        """
        CREATE TABLE IF NOT EXISTS events
        AS SELECT * FROM "%s"
        """
    db.execute(stmt % event_log_file).fetchall()

    print()

    rows = db.execute("""
        SELECT referrer, count(*)
        FROM events
        WHERE name = 'PageView'
        GROUP BY referrer
        ORDER BY count(*) DESC
    """).fetchall()
    print()
    print("Top Referrers")
    print(tabulate(rows,
                   headers=['referrer', 'page views'],
                   tablefmt="mixed_grid"))

    rows = db.execute("""
        SELECT sourceIp, count(*)
        FROM events
        WHERE name = 'PageView'
        GROUP BY sourceIp
        ORDER BY count(*) DESC
    """).fetchall()
    print()
    print("Most Frequent Visitors by IP")
    print(tabulate(rows,
                   headers=['sourceIp', 'page views'],
                   tablefmt="mixed_grid"))

    rows = db.execute("""
        SELECT pageUrl, count(*)
        FROM events
        GROUP BY pageUrl
        ORDER BY count(*) DESC
    """).fetchall()
    print()
    print("Most Viewed Pages")
    print(tabulate(rows,
                   headers=['pageUrl', 'page views'],
                   tablefmt="mixed_grid"))

    rows = db.execute("""
        SELECT time
        FROM events
        WHERE date_diff('day', current_date(), time) < 7
        AND name = 'PageView'
        ORDER BY time
    """).fetchall()

    counts_by_day = dp.Sequence(rows)\
        .map(lambda row: row[0])\
        .map(lambda d: d.strftime("%Y-%m-%d"))\
        .count(distinct=True)\
        .collect()

    counts_by_day = dict(counts_by_day)

    last_seven_days = dp.Sequence(range(7))\
        .map(lambda i: datetime.now(timezone.utc) - timedelta(days=i))\
        .map(lambda d: d.strftime("%Y-%m-%d"))\
        .map(lambda d: (d, counts_by_day.get(d, 0)))\
        .collect()

    print()
    print("Visitors Last 7 Days")
    print(tabulate(last_seven_days,
                   headers=['Date', 'Page Views'],
                   tablefmt="mixed_grid"))

    print()
    print()
    print()



import json
import gzip
from . import models
import pathlib
import flask
from flask import Flask, render_template, make_response, request
from base64 import b64decode
from datetime import datetime, timezone


# Setup
# http://probablyprogramming.com/2009/03/15/the-tiniest-gif-ever
# pixelGifBase64 = "R0lGODlhAQABAIAAAP///wAAACwAAAAAAQABAAACAkQBADs="
pixelGifBase64 = "R0lGODlhAQABAAD/ACwAAAAAAQABAAACADs="
pixelGif = b64decode(pixelGifBase64)
tracking_script = \
    """
    class Xem {

        constructor(siteId, xemUrl){
            this.xemUrl = xemUrl;
            this.siteId = siteId;
            this.currentUrl = "";
        }

        emit(event) {
            // encode data as base64 json string
            const payloadJSON = JSON.stringify(event);
            const payloadBase64 = btoa(payloadJSON);

            // build url
            const notificationUrl = this.xemUrl + "?" +
                "data=" + payloadBase64;

            console.log(notificationUrl)

            // send notificaiton to server
            let image = new Image();
            image.src = notificationUrl;
        }

        emitPageView() {
            const pageUrl = window.location.toString();
            const referrer = document.referrer;
            const event = {
                "name": "PageView",
                "siteId": this.siteId,
                "referrer": referrer,
                "pageUrl": pageUrl,
                "prevUrl": this.currentUrl
            }
            this.emit(event);
        }

        async start() {
            let newURL = window.location.toString();
            if (newURL !== this.currentURL) {
                this.emitPageView();
                this.currentURL = newURL;
            }
            setTimeout(() => {
                this.start()
            }, 1000);
        }
    }
    const xem = new Xem("%(siteId)s", "%(tracking_url)s");
    xem.start();
    """


# Flask App
template_folder = pathlib.Path(__file__)\
                         .expanduser()\
                         .parent\
                         .joinpath("templates")\
                         .resolve()
app = Flask(__name__, template_folder=template_folder)


# ======
# Routes
# ======

def home():
    return render_template('index.html')


def script():
    prop = request.args.get("siteId")
    root_url = request.url_root
    tracking_url = root_url + "tk"
    if prop is None:
        return ""

    params = {
        "tracking_url": tracking_url,
        "siteId": prop
    }
    response = make_response(tracking_script % params)
    response.headers['Content-Type'] = 'text/javascript'
    return response


def tracker():
    response = make_response(pixelGif)
    response.headers['Content-Type'] = 'image/gif'
    response.headers['Pragma'] = 'no-cache'
    response.headers['Expires'] = '-1'
    response.headers['Cache-Control'] = (
        'no-store, no-cache, must-revalidate,'
        ' post-check=0, pre-check=0,'
        ' max-age=0'
    )

    # begin tracking code
    event_data = read_event_data(request)
    event_source_ip = get_source_ip(request)
    event_time = datetime.now(timezone.utc)
    event_time_str = event_time.strftime("%Y-%m-%dT%H:%M:%SZ")

    # augment event data with time and ip
    event_data["time"] = event_time_str
    event_data["sourceIp"] = event_source_ip

    # validate event
    event = validate_event_data(event_data)

    # write to event log
    event_log = app.config["event_log"]
    with gzip.GzipFile(event_log, "a") as fh:
        event_json = event.model_dump_json()
        fh.write(bytes(event_json, "utf-8"))
        fh.write(bytes("\n", "utf-8"))

    return response


# =========
# Utilities
# =========

def read_event_data(req: flask.Request):
    event_base64 = req.args.get("data")
    if event_base64 is None:
        flask.abort(400, "no data sent")
    event_json = b64decode(event_base64.encode("ascii"))
    event_data = json.loads(event_json)
    return event_data


def get_source_ip(request: flask.Request):
    """
    Get source IP Address from request
    """
    if request.headers.getlist("X-Forwarded-For"):
        return request.headers.getlist("X-Forwarded-For")[0]
    else:
        return request.remote_addr


def validate_event_data(event_data: dict):
    """
    Validate user sent event data
    """
    etype = models.Event
    if event_data.get("name") == "PageView":
        etype = models.PageViewEvent
    return etype.model_validate(event_data)


# ===============
# Register routes
# ===============

app.route('/')(home)
app.route('/xem.js')(script)
app.route('/tk')(tracker)
# app.route('/test')(test)



from argparse import ArgumentParser
from .gapp import GunicornApplication
from .server import app
import multiprocessing as mp


if __name__ == '__main__':
    # Argument parser
    ap = ArgumentParser()
    ap.add_argument("-l", "--log", default="EventLog.jsonl.gz",
                    help="path to where we should save event log file")
    ap.add_argument("-p", "--port", default="3939",
                    help="port to listen for connections")
    ap.add_argument("-H", "--host", default="127.0.0.1",
                    help="host to listen for connections")
    ap.add_argument("-d", "--debug", action="store_true", default=False,
                    help="host to listen for connections")

    args = ap.parse_args()
    app.config["event_log"] = args.log

    if args.debug:
        app.run(debug=True, host=args.host, port=args.port)
    else:
        cpu_count = mp.cpu_count()
        options = {
            'bind': '%s:%s' % (args.host, args.port),
            'workers': min(cpu_count+1, 4),
            'timeout': 120,
        }
        guni_app = GunicornApplication(app, options)
        guni_app.run()
